 
import java.util.ArrayList;
/**
 * Verwaltet einen ungerichteten, gewichteten Graphen mittels Adjazenzmatrix
 * 
 * @author Albert Wiedemann 
 * @version 1.0
 */
class GraphMatrix
{
    /** Feld der Knoten des Graphen */
    private ArrayList<Knoten> knoten;   
    /** 2-dim Feld der Adjazenzmatrix */
    private ArrayList<ArrayList<Integer>> matrix;
    /** Feld der Kantensymbole des Graphen */
    private ArrayList<KantenSymbol> kanten;
    /** Die Warteliste (ToDo-Liste) */
    private ArrayList<Integer> warteliste;
    /** Feld der fertigen Knoten */
    private ArrayList<Integer> fertigeKnoten;
    
    /**
     * Baut die Datenstruktur auf
     */
    GraphMatrix ()
    {
        knoten = new ArrayList<Knoten>();
        matrix = new ArrayList<ArrayList<Integer>>();
        kanten = new ArrayList<KantenSymbol>();
        warteliste = new ArrayList<Integer>();
        fertigeKnoten = new ArrayList<Integer>();
    }
    
    /**
     * Einfügen eines neuen Knoten in den Graphen
     * @param bezeichner Bezeichner des neuen Knotens, der dem Graphen hinzugefügt wird.
     * @param x x-Koordinate für die Anzeige des Knotens
     * @param y y-Koordinate für die Anzeige des Knotens
     */
    void KnotenEinfügen(String bezeichner, int x, int y)
    {
        knoten.add(new Knoten(bezeichner, x, y));
        for (int index = 0; index < matrix.size(); index++)
        {
            matrix.get(index).add(new Integer(-1));
        }
        matrix.add(new ArrayList<Integer>());
        for (int index = 0; index < matrix.size(); index++)
        {
            matrix.get(matrix.size()-1).add(new Integer(-1));
        }
    }

    /**
     * Gibt den Knoten mit dem angegebenen Bezeichner zurück.
     * Wenn ein Knoten mit diesem Bezeichner nicht bekannt ist wird null zurückgegeben
     * @param bezeichner Bezeichner des Knoten der gesucht wird
     * @return Referenz auf das Knotenobjekt oder null
     */   
    private Knoten KnotenGeben(String bezeichner)
    {
        for (Knoten k: knoten)
        {
            if (k.BezeichnerGeben().equals(bezeichner))
            {
                return k;
            }
        }
        return null;
    }

    /**
     * Gibt die interne Nummer des Knotens
     * Wenn ein Knoten mit diesem Bezeichner nicht bekannt ist wird -1 zurückgegeben
     * @param bezeichner Bezeichner des Knoten der gesucht wird
     * @return Indexnummer des Knotens im Knotenfeld; 0 <= res <= knoten.size()-1 bzw. -1
     */   
    int KnotenNummerGeben(String bezeichner)
    {
        for (int index=0; index < knoten.size(); index++)
        {
            if (knoten.get(index).BezeichnerGeben().equals(bezeichner))
            {
                return index;
            }
        }
        return -1;
    }
    
    /**
     * Gibt die Bezeichnung eines Knotens mit der internen Knotennummer
     * @param Indexnummer des Knotens im Knotenarray; 0<= x <= knoten.size()
     * @return Bezeichner des Knotens
     */   
    String KnotenBezeichnerGeben(int knotenNummer)
    {
        if ((knotenNummer < knoten.size()) && (knotenNummer >= 0))
        {
            return knoten.get(knotenNummer).BezeichnerGeben();
        }
        else
        {
            return "";
        }
    }

    /**
     * Einfügen einer Kante in den Graphen
     * Eine Kante ist durch einen Anfangsknoten und einen Endknoten festgelegt und hat eine Gewichtung
     * @param von Bezeichner des Anfangsknotens
     * @param nach Bezeichner des Endknotens
     * @param gewichtung Gewichtung der Kante als Ganzzahl
     */
    void KanteEinfügen(String von, String nach, int gewichtung)
    {
        int vonNummer, nachNummer;
        vonNummer = KnotenNummerGeben(von);
        nachNummer = KnotenNummerGeben(nach);
        if ((vonNummer!=-1) && (nachNummer!=-1) && (vonNummer!=nachNummer))
        {
            matrix.get(vonNummer).set(nachNummer, gewichtung);
            matrix.get(nachNummer).set(vonNummer, gewichtung);
            kanten.add(new KantenSymbol(knoten.get(vonNummer).SymbolGeben(), knoten.get(nachNummer).SymbolGeben(), false, "" + gewichtung, 3, "blau"));
        }
    }

    /**
     * Gibt die Adjazenzmatrix des Graphen in der Konsole aus
     * Nach Zeilen und Spalten formatiert
     * Als Spaltenbreite wurde hier 4 Zeichen gewählt.
     */   
    void Ausgeben()
    {
        // Kopfzeile
        System.out.print("    ");  
        for (int i = 0; i < knoten.size(); i++)
        {
            System.out.print((knoten.get(i).BezeichnerGeben() + "    ").substring(0, 4));
        }
        System.out.println();
        
        for (int i = 0; i < knoten.size(); i++)
        {
            System.out.print((knoten.get(i).BezeichnerGeben() + "    ").substring(0, 4));
            for (int j = 0; j < knoten.size(); j++)
            {
                if (matrix.get(i).get(j) != -1)
                {
                    System.out.print((matrix.get(i).get(j)+ "   ").substring(0, 4));
                }
                else
                {
                    System.out.print("    ");
                }
            }
            System.out.println();           
        }
    }    

    /**
     * Gibt die Anzahl der Knoten des Graphen zurück
     * @return  Anzahl der Knoten
     */   
    int KnotenAnzahlgeben()
    {
        return knoten.size();
    }
    
    /**
     * Gibt die Gewichtung einer Kante zurück
     * Die Kante ist durch einen Anfangsknoten und einen Endknoten festgelegt
     * @param von Bezeichner des Anfangsknotens
     * @param nach Bezeichner des Endknotens
     * @return Gewichtung der Kante
     */ 
    int KanteGewichtGeben(String von, String nach)
    {
        int vonNummer, nachNummer;
        
        vonNummer = KnotenNummerGeben(von);
        nachNummer = KnotenNummerGeben(nach);
        if ((vonNummer!=-1) && (nachNummer!=-1))
        {
            return matrix.get(vonNummer).get(nachNummer);
        }
        else
        {
            return -1;
        }
    }
    
    /**
     * Löscht die Kanten und Knoten des Graphen
     * Die Anzeige wird auch gelöscht
     */
    void ZurückSetzen()
    {
        for (KantenSymbol k: kanten)
        {
            k.Entfernen();
        }
        for (Knoten k: knoten)
        {
            k.SymbolGeben().Entfernen();
        }
        knoten.clear();
        kanten.clear();
        matrix.clear();
    }
    
    /**
     * Führt den Dijkstra-Algorithmus vom Startknoten zum Zielknoten durch
     * @param startKnoten die Nummer des Startknotens
     * @param zielKnoten die Nummer des Zielknotens
     */
    void DijkstraAusführen(int startKnoten, int zielKnoten)
    {
        warteliste.clear();
        fertigeKnoten.clear();
        knoten.get(startKnoten).LängeSetzen(0);
        knoten.get(startKnoten).VorgängerSetzen(-1);
        int aktuellerKnoten = startKnoten;
        //Den Breitensuchealgorithmus gemäß der Teilaufgaben zum Dijkstra-Algorithmus ergänzen
        while ((aktuellerKnoten != zielKnoten) && (aktuellerKnoten != -1))
        {
            //Kantengewicht statt Kantenanzahl ergänzen
            for (int nummer = 0; nummer < knoten.size(); nummer += 1)
            {
                // Behandlung der Konten in der Warteliste ergänzen
                if ((matrix.get(aktuellerKnoten).get(nummer) > 0) && (! fertigeKnoten.contains(nummer)) && (! warteliste.contains(nummer)))
                {
                    warteliste.add(nummer);
                    knoten.get(nummer).VorgängerSetzen(aktuellerKnoten);
                    knoten.get(nummer).LängeSetzen(knoten.get(aktuellerKnoten).LängeGeben() + 1);
                }
            }
            fertigeKnoten.add(aktuellerKnoten);
            //Suche nach dem optimalen Knoten (mit dem bisher kürzesten Weg) ergänzen
            if (warteliste.size() > 0)
            {
                aktuellerKnoten = warteliste.get(0);
                warteliste.remove(0);
            }
            else
            {
                aktuellerKnoten = -1;
            }
        }
        if (aktuellerKnoten != -1)
        {
            System.out.println("Der Weg führt über " + knoten.get(aktuellerKnoten).LängeGeben() + " Kanten");
            ArrayList<String> namen = new ArrayList<String>();
            while (aktuellerKnoten != -1)
            {
                knoten.get(aktuellerKnoten).FarbeSetzen("grün");
                namen.add(knoten.get(aktuellerKnoten).BezeichnerGeben());
                aktuellerKnoten = knoten.get(aktuellerKnoten).VorgängerGeben();
            }
            for (int index = namen.size() - 1; index >= 0; index -= 1)
            {
                System.out.println(namen.get(index));
            }
        }
        else
        {
            System.out.println("Ziel nicht erreichbar");
        }
    }
}
